import { IQueryHandler, QueryHandler } from '@nestjs/cqrs';
import { MatchQuery, ScheduleItem } from './match.query';
import { Algorithm } from './algorithm.abstract';
import { PassengerOrientedAlgorithm } from './passenger-oriented-algorithm';
import { AlgorithmType } from '../../types/algorithm.types';
import { Inject } from '@nestjs/common';
import { AdRepositoryPort } from '@modules/ad/core/application/ports/ad.repository.port';
import {
  AD_REPOSITORY,
  INPUT_DATETIME_TRANSFORMER,
  MATCHING_REPOSITORY,
  PARAMS_PROVIDER,
} from '@modules/ad/ad.di-tokens';
import { MatchEntity } from '@modules/ad/core/domain/match.entity';
import { DefaultParamsProviderPort } from '../../ports/default-params-provider.port';
import { DefaultParams } from '../../ports/default-params.type';
import { DateTimeTransformerPort } from '../../ports/datetime-transformer.port';
import { Paginator } from '@mobicoop/ddd-library';
import { MatchingEntity } from '@modules/ad/core/domain/matching.entity';
import { MatchingRepositoryPort } from '../../ports/matching.repository.port';

@QueryHandler(MatchQuery)
export class MatchQueryHandler implements IQueryHandler {
  private readonly _defaultParams: DefaultParams;

  constructor(
    @Inject(PARAMS_PROVIDER)
    private readonly defaultParamsProvider: DefaultParamsProviderPort,
    @Inject(AD_REPOSITORY) private readonly adRepository: AdRepositoryPort,
    @Inject(MATCHING_REPOSITORY)
    private readonly matchingRepository: MatchingRepositoryPort,
    @Inject(INPUT_DATETIME_TRANSFORMER)
    private readonly datetimeTransformer: DateTimeTransformerPort,
  ) {
    this._defaultParams = defaultParamsProvider.getParams();
  }

  execute = async (query: MatchQuery): Promise<MatchingResult> => {
    query
      .setMissingMarginDurations(this._defaultParams.DEPARTURE_TIME_MARGIN)
      .setMissingStrict(this._defaultParams.STRICT)
      .setDefaultDriverAndPassengerParameters({
        driver: this._defaultParams.DRIVER,
        passenger: this._defaultParams.PASSENGER,
        seatsProposed: this._defaultParams.SEATS_PROPOSED,
        seatsRequested: this._defaultParams.SEATS_REQUESTED,
      })
      .setDefaultAlgorithmParameters({
        algorithmType: this._defaultParams.ALGORITHM_TYPE,
        remoteness: this._defaultParams.REMOTENESS,
        useProportion: this._defaultParams.USE_PROPORTION,
        proportion: this._defaultParams.PROPORTION,
        useAzimuth: this._defaultParams.USE_AZIMUTH,
        azimuthMargin: this._defaultParams.AZIMUTH_MARGIN,
        maxDetourDistanceRatio: this._defaultParams.MAX_DETOUR_DISTANCE_RATIO,
        maxDetourDurationRatio: this._defaultParams.MAX_DETOUR_DURATION_RATIO,
      })
      .setDefaultPagination({
        page: 1,
        perPage: this._defaultParams.PER_PAGE,
      })
      .setDatesAndSchedule(this.datetimeTransformer);
    let matchingEntity: MatchingEntity | undefined = await this._cachedMatching(
      query.id,
    );
    if (!matchingEntity)
      matchingEntity = (await this._createMatching(query)) as MatchingEntity;
    const perPage: number = query.perPage as number;
    const page: number = Paginator.pageNumber(
      matchingEntity.getProps().matches.length,
      perPage,
      query.page as number,
    );
    return {
      id: matchingEntity.id,
      matches: Paginator.pageItems(
        matchingEntity.getProps().matches,
        page,
        perPage,
      ),
      total: matchingEntity.getProps().matches.length,
      page,
      perPage,
    };
  };

  private _cachedMatching = async (
    id?: string,
  ): Promise<MatchingEntity | undefined> => {
    if (!id) return undefined;
    try {
      return await this.matchingRepository.get(id);
    } catch (e: any) {
      return undefined;
    }
  };

  private _createMatching = async (
    query: MatchQuery,
  ): Promise<MatchingEntity> => {
    await query.setRoutes();

    let algorithm: Algorithm;
    switch (query.algorithmType) {
      case AlgorithmType.PASSENGER_ORIENTED:
      default:
        algorithm = new PassengerOrientedAlgorithm(query, this.adRepository);
    }

    const matches: MatchEntity[] = await algorithm.match();
    const matchingEntity = MatchingEntity.create({
      matches,
      query: {
        driver: query.driver as boolean,
        passenger: query.passenger as boolean,
        frequency: query.frequency,
        fromDate: query.fromDate,
        toDate: query.toDate,
        schedule: query.schedule.map((scheduleItem: ScheduleItem) => ({
          day: scheduleItem.day as number,
          time: scheduleItem.time,
          margin: scheduleItem.margin as number,
        })),
        seatsProposed: query.seatsProposed as number,
        seatsRequested: query.seatsRequested as number,
        strict: query.strict as boolean,
        waypoints: query.waypoints,
        algorithmType: query.algorithmType as AlgorithmType,
        remoteness: query.remoteness as number,
        useProportion: query.useProportion as boolean,
        proportion: query.proportion as number,
        useAzimuth: query.useAzimuth as boolean,
        azimuthMargin: query.azimuthMargin as number,
        maxDetourDistanceRatio: query.maxDetourDistanceRatio as number,
        maxDetourDurationRatio: query.maxDetourDurationRatio as number,
      },
    });
    await this.matchingRepository.save(matchingEntity);
    return matchingEntity;
  };
}

export type MatchingResult = {
  id: string;
  matches: MatchEntity[];
  total: number;
  page: number;
  perPage: number;
};
