import { ArgumentInvalidException, ValueObject } from '@mobicoop/ddd-library';
import { JourneyItem, JourneyItemProps } from './journey-item.value-object';
import { ActorTime } from './actor-time.value-object';
import { Role } from '../ad.types';
import { Target } from '../candidate.types';
import { Point } from './point.value-object';

/** Note:
 * Value Objects with multiple properties can contain
 * other Value Objects inside if needed.
 * */

export interface JourneyProps {
  firstDate: Date;
  lastDate: Date;
  journeyItems: JourneyItemProps[];
}

export class Journey extends ValueObject<JourneyProps> {
  get firstDate(): Date {
    return this.props.firstDate;
  }

  get lastDate(): Date {
    return this.props.lastDate;
  }

  get journeyItems(): JourneyItemProps[] {
    return this.props.journeyItems;
  }

  hasValidPickUp = (): boolean => {
    const passengerDepartureJourneyItem: JourneyItem = this.journeyItems.find(
      (journeyItem: JourneyItem) =>
        journeyItem.actorTimes.find(
          (actorTime: ActorTime) =>
            actorTime.role == Role.PASSENGER &&
            actorTime.target == Target.START,
        ) as ActorTime,
    ) as JourneyItem;
    const passengerDepartureActorTime =
      passengerDepartureJourneyItem.actorTimes.find(
        (actorTime: ActorTime) =>
          actorTime.role == Role.PASSENGER && actorTime.target == Target.START,
      ) as ActorTime;
    const driverNeutralActorTime =
      passengerDepartureJourneyItem.actorTimes.find(
        (actorTime: ActorTime) =>
          actorTime.role == Role.DRIVER && actorTime.target == Target.NEUTRAL,
      ) as ActorTime;
    return (
      (passengerDepartureActorTime.firstMinDatetime <=
        driverNeutralActorTime.firstMaxDatetime &&
        driverNeutralActorTime.firstMaxDatetime <=
          passengerDepartureActorTime.firstMaxDatetime) ||
      (passengerDepartureActorTime.firstMinDatetime <=
        driverNeutralActorTime.firstMinDatetime &&
        driverNeutralActorTime.firstMinDatetime <=
          passengerDepartureActorTime.firstMaxDatetime)
    );
  };

  firstDriverDepartureTime = (): string => {
    const firstDriverDepartureDatetime: Date = (
      this._driverDepartureJourneyItem().actorTimes.find(
        (actorTime: ActorTime) =>
          actorTime.role == Role.DRIVER && actorTime.target == Target.START,
      ) as ActorTime
    ).firstDatetime;
    return `${firstDriverDepartureDatetime
      .getUTCHours()
      .toString()
      .padStart(2, '0')}:${firstDriverDepartureDatetime
      .getUTCMinutes()
      .toString()
      .padStart(2, '0')}`;
  };

  driverOrigin = (): Point =>
    new Point({
      lon: this._driverDepartureJourneyItem().lon,
      lat: this._driverDepartureJourneyItem().lat,
    });

  private _driverDepartureJourneyItem = (): JourneyItem =>
    this.journeyItems.find(
      (journeyItem: JourneyItem) =>
        journeyItem.actorTimes.find(
          (actorTime: ActorTime) =>
            actorTime.role == Role.DRIVER && actorTime.target == Target.START,
        ) as ActorTime,
    ) as JourneyItem;

  protected validate(props: JourneyProps): void {
    if (props.firstDate.getUTCDay() != props.lastDate.getUTCDay())
      throw new ArgumentInvalidException(
        'firstDate week day must be equal to lastDate week day',
      );
    if (props.firstDate > props.lastDate)
      throw new ArgumentInvalidException('firstDate must be before lastDate');
    if (props.journeyItems.length < 2)
      throw new ArgumentInvalidException(
        'at least 2 journey items are required',
      );
  }
}
