import {
  CalendarTools,
  CalendarToolsException,
} from '@modules/ad/core/domain/calendar-tools.service';

describe('Calendar tools service', () => {
  describe('First date', () => {
    it('should return the first date for a given week day within a date range', () => {
      const firstDate: Date = CalendarTools.firstDate(1, {
        lowerDate: '2023-08-31',
        higherDate: '2023-09-07',
      });
      expect(firstDate.getUTCDay()).toBe(1);
      expect(firstDate.getUTCDate()).toBe(4);
      expect(firstDate.getUTCMonth()).toBe(8);
      const secondDate: Date = CalendarTools.firstDate(5, {
        lowerDate: '2023-08-31',
        higherDate: '2023-09-07',
      });
      expect(secondDate.getUTCDay()).toBe(5);
      expect(secondDate.getUTCDate()).toBe(1);
      expect(secondDate.getUTCMonth()).toBe(8);
      const thirdDate: Date = CalendarTools.firstDate(4, {
        lowerDate: '2023-08-31',
        higherDate: '2023-09-07',
      });
      expect(thirdDate.getUTCDay()).toBe(4);
      expect(thirdDate.getUTCDate()).toBe(31);
      expect(thirdDate.getUTCMonth()).toBe(7);
    });
    it('should throw an exception if a given week day is not within a date range', () => {
      expect(() => {
        CalendarTools.firstDate(1, {
          lowerDate: '2023-09-05',
          higherDate: '2023-09-07',
        });
      }).toThrow(CalendarToolsException);
    });
    it('should throw an exception if a given week day is invalid', () => {
      expect(() => {
        CalendarTools.firstDate(8, {
          lowerDate: '2023-09-05',
          higherDate: '2023-09-07',
        });
      }).toThrow(CalendarToolsException);
    });
  });

  describe('Second date', () => {
    it('should return the last date for a given week day within a date range', () => {
      const firstDate: Date = CalendarTools.lastDate(0, {
        lowerDate: '2023-09-30',
        higherDate: '2024-09-30',
      });
      expect(firstDate.getUTCDay()).toBe(0);
      expect(firstDate.getUTCDate()).toBe(29);
      expect(firstDate.getUTCMonth()).toBe(8);
      const secondDate: Date = CalendarTools.lastDate(5, {
        lowerDate: '2023-09-30',
        higherDate: '2024-09-30',
      });
      expect(secondDate.getUTCDay()).toBe(5);
      expect(secondDate.getUTCDate()).toBe(27);
      expect(secondDate.getUTCMonth()).toBe(8);
      const thirdDate: Date = CalendarTools.lastDate(1, {
        lowerDate: '2023-09-30',
        higherDate: '2024-09-30',
      });
      expect(thirdDate.getUTCDay()).toBe(1);
      expect(thirdDate.getUTCDate()).toBe(30);
      expect(thirdDate.getUTCMonth()).toBe(8);
    });
    it('should throw an exception if a given week day is not within a date range', () => {
      expect(() => {
        CalendarTools.lastDate(2, {
          lowerDate: '2024-09-27',
          higherDate: '2024-09-30',
        });
      }).toThrow(CalendarToolsException);
    });
    it('should throw an exception if a given week day is invalid', () => {
      expect(() => {
        CalendarTools.lastDate(8, {
          lowerDate: '2023-09-30',
          higherDate: '2024-09-30',
        });
      }).toThrow(CalendarToolsException);
    });
  });

  describe('Datetime from string', () => {
    it('should return a date with time from a string without additional seconds', () => {
      const datetime: Date = CalendarTools.datetimeWithSeconds(
        new Date('2023-09-01'),
        '07:12',
      );
      expect(datetime.getUTCMinutes()).toBe(12);
    });
    it('should return a date with time from a string with additional seconds', () => {
      const datetime: Date = CalendarTools.datetimeWithSeconds(
        new Date('2023-09-01'),
        '07:12',
        60,
      );
      expect(datetime.getUTCMinutes()).toBe(13);
    });
    it('should return a date with time from a string with negative additional seconds', () => {
      const datetime: Date = CalendarTools.datetimeWithSeconds(
        new Date('2023-09-01'),
        '07:00',
        -60,
      );
      expect(datetime.getUTCHours()).toBe(6);
      expect(datetime.getUTCMinutes()).toBe(59);
    });
  });

  describe('epochDaysFromTime', () => {
    it('should return the epoch day for day 1', () => {
      const days: Date[] = CalendarTools.epochDaysFromTime(1, '07:00');
      expect(days).toHaveLength(1);
      expect(days[0].getUTCFullYear()).toBe(1969);
      expect(days[0].getUTCMonth()).toBe(11);
      expect(days[0].getUTCDate()).toBe(29);
    });
    it('should return the epoch day for day 2', () => {
      const days: Date[] = CalendarTools.epochDaysFromTime(2, '07:00');
      expect(days).toHaveLength(1);
      expect(days[0].getUTCFullYear()).toBe(1969);
      expect(days[0].getUTCMonth()).toBe(11);
      expect(days[0].getUTCDate()).toBe(30);
    });
    it('should return the epoch day for day 3', () => {
      const days: Date[] = CalendarTools.epochDaysFromTime(3, '07:00');
      expect(days).toHaveLength(1);
      expect(days[0].getUTCFullYear()).toBe(1969);
      expect(days[0].getUTCMonth()).toBe(11);
      expect(days[0].getUTCDate()).toBe(31);
    });
    it('should return the epoch day for day 4', () => {
      const days: Date[] = CalendarTools.epochDaysFromTime(4, '07:00');
      expect(days).toHaveLength(1);
      expect(days[0].getUTCFullYear()).toBe(1970);
      expect(days[0].getUTCMonth()).toBe(0);
      expect(days[0].getUTCDate()).toBe(1);
    });
    it('should return the epoch day for day 5', () => {
      const days: Date[] = CalendarTools.epochDaysFromTime(5, '07:00');
      expect(days).toHaveLength(1);
      expect(days[0].getUTCFullYear()).toBe(1970);
      expect(days[0].getUTCMonth()).toBe(0);
      expect(days[0].getUTCDate()).toBe(2);
    });
    it('should return the epoch days for day 0', () => {
      const days: Date[] = CalendarTools.epochDaysFromTime(0, '07:00');
      expect(days).toHaveLength(2);
      expect(days[0].getUTCFullYear()).toBe(1969);
      expect(days[0].getUTCMonth()).toBe(11);
      expect(days[0].getUTCDate()).toBe(28);
      expect(days[1].getUTCFullYear()).toBe(1970);
      expect(days[1].getUTCMonth()).toBe(0);
      expect(days[1].getUTCDate()).toBe(4);
    });
    it('should return the epoch days for day 6', () => {
      const days: Date[] = CalendarTools.epochDaysFromTime(6, '07:00');
      expect(days).toHaveLength(2);
      expect(days[0].getUTCFullYear()).toBe(1969);
      expect(days[0].getUTCMonth()).toBe(11);
      expect(days[0].getUTCDate()).toBe(27);
      expect(days[1].getUTCFullYear()).toBe(1970);
      expect(days[1].getUTCMonth()).toBe(0);
      expect(days[1].getUTCDate()).toBe(3);
    });
    it('should throw an exception if a given week day is invalid', () => {
      expect(() => {
        CalendarTools.epochDaysFromTime(8, '07:00');
      }).toThrow(CalendarToolsException);
    });
  });
});
