import {
  AD_REPOSITORY,
  INPUT_DATETIME_TRANSFORMER,
  MATCHING_REPOSITORY,
  PARAMS_PROVIDER,
} from '@modules/ad/ad.di-tokens';
import { DateTimeTransformerPort } from '@modules/ad/core/application/ports/datetime-transformer.port';
import { DefaultParamsProviderPort } from '@modules/ad/core/application/ports/default-params-provider.port';
import { MatchingRepositoryPort } from '@modules/ad/core/application/ports/matching.repository.port';
import { RouteProviderPort } from '@modules/ad/core/application/ports/route-provider.port';
import { MatchQuery } from '@modules/ad/core/application/queries/match/match.query';
import {
  MatchQueryHandler,
  MatchingResult,
} from '@modules/ad/core/application/queries/match/match.query-handler';
import { AlgorithmType } from '@modules/ad/core/application/types/algorithm.types';
import { Waypoint } from '@modules/ad/core/application/types/waypoint.type';
import { Frequency, Role } from '@modules/ad/core/domain/ad.types';
import { Target } from '@modules/ad/core/domain/candidate.types';
import { MatchEntity } from '@modules/ad/core/domain/match.entity';
import { MatchingEntity } from '@modules/ad/core/domain/matching.entity';
import { Test, TestingModule } from '@nestjs/testing';

const originWaypoint: Waypoint = {
  position: 0,
  lat: 48.689445,
  lon: 6.17651,
  houseNumber: '5',
  street: 'Avenue Foch',
  locality: 'Nancy',
  postalCode: '54000',
  country: 'France',
};
const destinationWaypoint: Waypoint = {
  position: 1,
  lat: 48.8566,
  lon: 2.3522,
  locality: 'Paris',
  postalCode: '75000',
  country: 'France',
};

const mockAdRepository = {
  getCandidateAds: jest.fn().mockImplementation(() => [
    {
      id: 'cc260669-1c6d-441f-80a5-19cd59afb777',
      getProps: jest.fn().mockImplementation(() => ({
        role: Role.DRIVER,
        waypoints: [
          {
            lat: 48.68787,
            lon: 6.165871,
          },
          {
            lat: 48.97878,
            lon: 2.45787,
          },
        ],
      })),
    },
    {
      id: '4431adea-2e10-4032-a743-01d537058914',
      getProps: jest.fn().mockImplementation(() => ({
        role: Role.DRIVER,
        waypoints: [
          {
            lat: 48.698754,
            lon: 6.159874,
          },
          {
            lat: 48.969874,
            lon: 2.449875,
          },
        ],
      })),
    },
  ]),
};

const mockMatchingRepository: MatchingRepositoryPort = {
  get: jest
    .fn()
    .mockImplementationOnce(
      () =>
        new MatchingEntity({
          id: 'a3b10efb-121e-4d08-9198-9f57afdb5e2d',
          createdAt: new Date('2023-08-20T09:48:00Z'),
          updatedAt: new Date('2023-08-20T09:48:00Z'),
          props: {
            matches: [
              new MatchEntity({
                id: '4bd4e90b-ffba-4f5f-b904-48ad0667a1d7',
                createdAt: new Date('2023-08-30T08:45:00Z'),
                updatedAt: new Date('2023-08-30T08:45:00Z'),
                props: {
                  adId: 'dd937edf-1264-4868-b073-d1952abe30b1',
                  role: Role.DRIVER,
                  frequency: Frequency.PUNCTUAL,
                  distance: 356041,
                  duration: 12647,
                  initialDistance: 348745,
                  initialDuration: 12105,
                  distanceDetour: 7296,
                  durationDetour: 542,
                  distanceDetourPercentage: 4.1,
                  durationDetourPercentage: 3.8,
                  journeys: [
                    {
                      firstDate: new Date('2023-08-28'),
                      lastDate: new Date('2023-08-28'),
                      journeyItems: [
                        {
                          lon: 6.389745,
                          lat: 48.32644,
                          duration: 0,
                          distance: 0,
                          actorTimes: [
                            {
                              role: Role.DRIVER,
                              target: Target.START,
                              firstDatetime: new Date('2023-08-28T07:00:00Z'),
                              firstMinDatetime: new Date(
                                '2023-08-28T06:45:00Z',
                              ),
                              firstMaxDatetime: new Date(
                                '2023-08-28T07:15:00Z',
                              ),
                              lastDatetime: new Date('2023-08-28T07:00:00Z'),
                              lastMinDatetime: new Date('2023-08-28T06:45:00Z'),
                              lastMaxDatetime: new Date('2023-08-28T07:15:00Z'),
                            },
                            {
                              role: Role.PASSENGER,
                              target: Target.START,
                              firstDatetime: new Date('2023-08-28T07:00:00Z'),
                              firstMinDatetime: new Date(
                                '2023-08-28T06:45:00Z',
                              ),
                              firstMaxDatetime: new Date(
                                '2023-08-28T07:15:00Z',
                              ),
                              lastDatetime: new Date('2023-08-28T07:00:00Z'),
                              lastMinDatetime: new Date('2023-08-28T06:45:00Z'),
                              lastMaxDatetime: new Date('2023-08-28T07:15:00Z'),
                            },
                          ],
                        },
                        {
                          lon: 6.984567,
                          lat: 48.021548,
                          distance: 356041,
                          duration: 12647,
                          actorTimes: [
                            {
                              role: Role.DRIVER,
                              target: Target.FINISH,
                              firstDatetime: new Date('2023-08-28T07:00:00Z'),
                              firstMinDatetime: new Date(
                                '2023-08-28T06:45:00Z',
                              ),
                              firstMaxDatetime: new Date(
                                '2023-08-28T07:15:00Z',
                              ),
                              lastDatetime: new Date('2023-08-28T07:00:00Z'),
                              lastMinDatetime: new Date('2023-08-28T06:45:00Z'),
                              lastMaxDatetime: new Date('2023-08-28T07:15:00Z'),
                            },
                            {
                              role: Role.PASSENGER,
                              target: Target.FINISH,
                              firstDatetime: new Date('2023-08-28T07:00:00Z'),
                              firstMinDatetime: new Date(
                                '2023-08-28T06:45:00Z',
                              ),
                              firstMaxDatetime: new Date(
                                '2023-08-28T07:15:00Z',
                              ),
                              lastDatetime: new Date('2023-08-28T07:00:00Z'),
                              lastMinDatetime: new Date('2023-08-28T06:45:00Z'),
                              lastMaxDatetime: new Date('2023-08-28T07:15:00Z'),
                            },
                          ],
                        },
                      ],
                    },
                  ],
                },
              }),
            ],
            query: {
              driver: false,
              passenger: true,
              frequency: Frequency.PUNCTUAL,
              fromDate: '2023-08-28',
              toDate: '2023-08-28',
              schedule: [
                {
                  day: 1,
                  time: '06:40',
                  margin: 900,
                },
              ],
              seatsProposed: 3,
              seatsRequested: 1,
              strict: true,
              waypoints: [
                {
                  lon: 6.389745,
                  lat: 48.32644,
                },
                {
                  lon: 6.984567,
                  lat: 48.021548,
                },
              ],
              algorithmType: 'PASSENGER_ORIENTED',
              remoteness: 15000,
              useProportion: true,
              proportion: 0.3,
              useAzimuth: true,
              azimuthMargin: 10,
              maxDetourDistanceRatio: 0.3,
              maxDetourDurationRatio: 0.3,
            },
          },
        }),
    )
    .mockImplementationOnce(() => {
      throw new Error();
    }),
  save: jest.fn(),
};

const mockDefaultParamsProvider: DefaultParamsProviderPort = {
  getParams: () => {
    return {
      DEPARTURE_TIME_MARGIN: 900,
      DRIVER: false,
      SEATS_PROPOSED: 3,
      PASSENGER: true,
      SEATS_REQUESTED: 1,
      STRICT: false,
      TIMEZONE: 'Europe/Paris',
      ALGORITHM_TYPE: AlgorithmType.PASSENGER_ORIENTED,
      REMOTENESS: 15000,
      USE_PROPORTION: true,
      PROPORTION: 0.3,
      USE_AZIMUTH: true,
      AZIMUTH_MARGIN: 10,
      MAX_DETOUR_DISTANCE_RATIO: 0.3,
      MAX_DETOUR_DURATION_RATIO: 0.3,
      PER_PAGE: 10,
    };
  },
};

const mockInputDateTimeTransformer: DateTimeTransformerPort = {
  fromDate: jest.fn(),
  toDate: jest.fn(),
  day: jest.fn(),
  time: jest.fn(),
};

const mockRouteProvider: RouteProviderPort = {
  getBasic: jest.fn().mockImplementation(() => ({
    distance: 350101,
    duration: 14422,
    fwdAzimuth: 273,
    backAzimuth: 93,
    distanceAzimuth: 336544,
    points: [],
  })),
  getDetailed: jest.fn(),
};

describe('Match Query Handler', () => {
  let matchQueryHandler: MatchQueryHandler;

  beforeAll(async () => {
    const module: TestingModule = await Test.createTestingModule({
      providers: [
        MatchQueryHandler,
        {
          provide: AD_REPOSITORY,
          useValue: mockAdRepository,
        },
        {
          provide: MATCHING_REPOSITORY,
          useValue: mockMatchingRepository,
        },
        {
          provide: PARAMS_PROVIDER,
          useValue: mockDefaultParamsProvider,
        },
        {
          provide: INPUT_DATETIME_TRANSFORMER,
          useValue: mockInputDateTimeTransformer,
        },
      ],
    }).compile();

    matchQueryHandler = module.get<MatchQueryHandler>(MatchQueryHandler);
  });

  afterEach(async () => {
    jest.clearAllMocks();
  });

  it('should be defined', () => {
    expect(matchQueryHandler).toBeDefined();
  });

  it('should return a Matching', async () => {
    jest.spyOn(MatchingEntity, 'create');
    const matchQuery = new MatchQuery(
      {
        algorithmType: AlgorithmType.PASSENGER_ORIENTED,
        driver: false,
        passenger: true,
        frequency: Frequency.PUNCTUAL,
        fromDate: '2023-08-28',
        toDate: '2023-08-28',
        schedule: [
          {
            time: '07:05',
            day: 1,
            margin: 900,
          },
        ],
        strict: false,
        waypoints: [originWaypoint, destinationWaypoint],
      },
      mockRouteProvider,
    );
    const matching: MatchingResult = await matchQueryHandler.execute(
      matchQuery,
    );
    expect(matching.id).toHaveLength(36);
    expect(MatchingEntity.create).toHaveBeenCalledTimes(1);
  });

  it('should return a valid saved Matching', async () => {
    jest.spyOn(MatchingEntity, 'create');
    const matchQuery = new MatchQuery(
      {
        id: 'a3b10efb-121e-4d08-9198-9f57afdb5e2d',
        algorithmType: AlgorithmType.PASSENGER_ORIENTED,
        driver: false,
        passenger: true,
        frequency: Frequency.PUNCTUAL,
        fromDate: '2023-08-28',
        toDate: '2023-08-28',
        schedule: [
          {
            time: '07:05',
            day: 1,
            margin: 900,
          },
        ],
        strict: false,
        waypoints: [originWaypoint, destinationWaypoint],
      },
      mockRouteProvider,
    );
    const matching: MatchingResult = await matchQueryHandler.execute(
      matchQuery,
    );
    expect(matching.id).toBe('a3b10efb-121e-4d08-9198-9f57afdb5e2d');
    expect(MatchingEntity.create).toHaveBeenCalledTimes(0);
  });

  it('should return a new matching if saved Matching is not found', async () => {
    jest.spyOn(MatchingEntity, 'create');
    const matchQuery = new MatchQuery(
      {
        id: 'a3b10efb-121e-4d08-9198-9f57afdb5e2d',
        algorithmType: AlgorithmType.PASSENGER_ORIENTED,
        driver: false,
        passenger: true,
        frequency: Frequency.PUNCTUAL,
        fromDate: '2023-08-28',
        toDate: '2023-08-28',
        schedule: [
          {
            time: '07:05',
            day: 1,
            margin: 900,
          },
        ],
        strict: false,
        waypoints: [originWaypoint, destinationWaypoint],
      },
      mockRouteProvider,
    );
    const matching: MatchingResult = await matchQueryHandler.execute(
      matchQuery,
    );
    expect(matching.id).toHaveLength(36);
    expect(MatchingEntity.create).toHaveBeenCalledTimes(1);
  });
});
